/* simplesynth.cc
 * ------------------------------------------------------------------------
 *
 * Copyright (C) 2000 Matthew Stanton <matthews@softhome.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * ------------------------------------------------------------------------
 *
 * A simple 2-oscillator virtual analogue synthesizer for OCTAL.
 *
 * Filter formulae are from Robert Bristow-Johnson's Audio EQ Cookbook.
 * <robert@audioheads.com>
 *
 * Oscillator::fast_sin is from David A. Bartold's analogue.cpp in the
 * Computer Music Toolkit library of LADSPA plugins.
 *
 * Some parameter scaling functions are from Makk's M4 synthesizer for Buzz.
 * <makk@gmx.de>
 *
 */

#define OX_CLASS SimpleSynth
#include "ox_wrappers.h"
#include "simplesynth.h"
#include "ox_wrappers.cc"

/********************************
  Oscillator
********************************/

Oscillator::Oscillator() {
	;
}

Oscillator::~Oscillator() {
	;
}

inline samp Oscillator::fast_sin(samp x) {
	if (x > PI)
		x = (x < PI * 1.5) ? (PI - x) : (x - 2.0 * PI);
	else if (x > PI * 0.5)
		x = PI - x;
	
	return (samp)(x * (1.05 - x * x * 0.175));
}

void Oscillator::set_wave(int w) {
	switch(w) {
	case 0:
		waveform = saw;
		break;
		
	case 1:
		waveform = pulse;
		break;
		
	case 2:
		waveform = triangle;
		break;
		
	case 3:
		waveform = sine;
		break;	
	}
}

char* Oscillator::get_wave_name(int w) {
	static char *wave_options[] = {"saw", "pulse", "triangle", "sine"};
	if (w > 3) return "OUT OF RANGE";
	return wave_options[w];
}

void Oscillator::set_pitch(freq p) {
	pitch = p;
	update_pitch();
}

inline void Oscillator::set_detune_ratio(freq d) {
	detune_ratio = d;
	update_pitch();
}

inline samp Oscillator::next_samp() {
	samp out;

	switch(waveform) {
	case saw:
		out = (samp)(phase - 0.5) * -2.0;
		break;
		
	case pulse:
		out = (samp)(phase >= 0.5) ? 1.0 : -1.0;
		break;
		
	case triangle:
		if (phase < 0.5) out = (samp)(phase - 0.25) * 4.0;
		else out = (samp)(phase - 0.75) * -4.0;
		break;
		
	case sine:
		out = fast_sin((samp)phase * 2 * PI);
		break;
		
	default:
		out = 0.0;
	}
	
	phase += phaseinc;
	if (phase >= 1.0) phase -= 1.0;
	
	return out;
}

inline void Oscillator::update_pitch() {
	phaseinc = (pitch * detune_ratio) / SAMPLE_RATE;
}


/********************************
  Envelope
********************************/

Envelope::Envelope() {
	stage = off;
	level = 0.0;
}

Envelope::~Envelope() {
	;
}

void Envelope::set_attack(samp a) {
	a_step = value_to_step(a);
}

void Envelope::set_decay(samp d) {
	d_step = value_to_step(d);
}

void Envelope::set_sustain(samp s) {
	s_level = s;
}

void Envelope::set_release(samp r) {
	r_step = value_to_step(r);
}

void Envelope::gate_on() {
	stage = attack;
	level = 0.0;
}

void Envelope::gate_off() {
	stage = release;
}

int Envelope::get_stage() {
	return (int)stage;
}

inline samp Envelope::next_samp() {
	// 0.0 is fully closed, 1.0 is fully open
	switch(stage) {
	case off:
		level = 0;
		break;
		
	case attack:
		level += a_step;
		if (level >= 1.0) {
			level = 1.0;
			stage = decay;
		}
		break;
		
	case decay:
		level -= d_step * (1.0 - s_level);
		if (level <= s_level) {
			level = s_level;
			stage = sustain;
		}
		break;
		
	case sustain:
		level = s_level;
		break;
		
	case release:
		level -= r_step;
		if (level <= 0.0) {
			level = 0;
			stage = off;
		}
		break;
		
	default:
		level = 0;
	}
	
	return level;
}

inline samp Envelope::value_to_time(samp x) {
	// return (pow(10000.0, x)/1000.0);
	if (x < 0) x = 0;
	if (x > 1) x = 1;
	return (pow(x, 2)*9.999 + .001);
}

inline samp Envelope::value_to_step(samp x) {
/*
	returns the amount the envelope needs to be incremented by in a single
	sample to fully open in a certain amount of time.  0 corresponds to 1ms,
	1 corresponds to 10 seconds.
*/
	return (samp)( 1.0 / (value_to_time(x) * SAMPLE_RATE));
}



/********************************
  Filter
********************************/

Filter::Filter() {
	coefs = new samp[5];
	old_in = new samp[2];
	old_out = new samp[2];
	clear();
}

Filter::~Filter() {
	delete coefs;
	delete old_in;
	delete old_out;
}

void Filter::set_cutoff(freq c) {
	cutoff = c;
}

void Filter::set_cutoff_offset(freq c) {
	cutoff_offset = c;
}

void Filter::set_resonance(freq r) {
	resonance = r;
}

void Filter::set_type(int t) {
	switch(t) {
	case 0:
		type = lowpass;
		break;
		
	case 1:
		type = highpass;
		break;
		
	case 2:
		type = bandpass;
		break;
		
	case 3:
		type = bandreject;
		break;
	}
}

char* Filter::get_type_name(int t) {
	static char* type_options[] = {"lowpass", "highpass", "bandpass", "bandreject"};
	if (t > 3) return "OUT OF RANGE";
	return type_options[t];
}

void Filter::compute_coefs() {
	// might want to make this static and use it to generate a table from
	// SimpleSynth::initialize()
	samp omega = 2 * PI * ((samp)(cutoff + cutoff_offset) / SAMPLE_RATE);
	samp sn = (samp)sin(omega);
	samp cs = (samp)cos(omega);
	samp alpha;
	
	if ((type == lowpass) || (type == highpass))
		alpha = sn / (2 * res_to_q(resonance));
	else if ((type == bandpass) || (type == bandreject))
		alpha = sn * (samp)sinh((log(2)/2) * res_to_bw(resonance) * (omega/sn));
	else
		alpha = 0;
	
	samp a0, a1, a2, b0, b1, b2;
	
	switch(type) {
	case lowpass:
		b0 = (1 - cs)/2;
		b1 = 1 - cs;
		b2 = (1 - cs)/2;
		a0 = 1 + alpha;
		a1 = -2 * cs;
		a2 = 1 - alpha;
		break;
		
	case highpass:
		b0 = (1 + cs)/2;
		b1 = -(1 + cs);
		b2 = (1 + cs)/2;
		a0 = 1 + alpha;
		a1 = -2 * cs;
		a2 = 1 - alpha;
		break;
		
	case bandpass:
		b0 = alpha;
		b1 = 0;
		b2 = -alpha;
		a0 = 1 + alpha;
		a1 = -2 * cs;
		a2 = 1 - alpha;
		break;
		
	case bandreject:
		b0 = 1;
		b1 = -2 * cs;
		b2 = 1;
		a0 = 1 + alpha;
		a1 = -2 * cs;
		a2 = 1 - alpha;
		break;
	
	default:
		b0 = 0;
		b1 = 0;
		b2 = 0;
		a0 = 1;
		a1 = 0;
		a2 = 0;
	}
	
	coefs[0] = b0 / a0;
	coefs[1] = b1 / a0;
	coefs[2] = b2 / a0;
	coefs[3] = a1 / a0;
	coefs[4] = a2 / a0;
	
}

inline samp Filter::next_samp(samp in) {
	samp out = coefs[0]*in + coefs[1]*old_in[0] + coefs[2]*old_in[1] - coefs[3]*old_out[0] - coefs[4]*old_out[1];
	old_in[1] = old_in[0];
	old_in[0] = in;
	old_out[1] = old_out[0];
	old_out[0] = out;
	// if (out >  .00000001) old_out[0] = out;
	// else old_out[0] = 0.0;
	return out;
}

inline samp Filter::res_to_q(freq r) {
	return (samp)(pow(r, 4)*150+0.1);
}

inline samp Filter::res_to_bw(freq r) {
	return (samp)(pow(r, 4)*4+0.1);
}

void Filter::clear() {
	old_in[0] = old_in[1] = old_out[0] = old_out [1] = 0.0;
}


/********************************
  SimpleSynth
********************************/

SimpleSynth::SimpleSynth(machine* m) : OX_Machine(m) {
	vol = 0x80;
	gate = 0;
}

SimpleSynth::~SimpleSynth() {
	;
}

int SimpleSynth::initialize(machine_type* t) {

	static param_spec my_params[] = {
		{
			OX_FORMAT_NOTE,
			OX_WIDGET_SLIDER,
			"note",
			"Which note to play.",
			0,
			NOTE_RANGE,
			0
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"vol",
			"Volume: 00-FF",
			0x00,
			0xFF,
			0x80
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_OPTION,
			"osc1 wave",
			"Osc 1 waveform",
			0,
			3,
			0
		},
			
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_OPTION,
			"osc2 wave",
			"Osc 2 waveform",
			0,
			3,
			0
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"osc1 detune",
			"Osc 1 detune: 00-FE",
			0x00,
			0xC8,
			0x64
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"osc2 detune",
			"Osc 2 detune: 00-FE",
			0x00,
			0xC8,
			0x64
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"osc mix",
			"Osc 1 / Osc 2 balance",
			0x00,
			0xFF,
			0x00
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"amp attack",
			"Amp env attack time",
			0x00,
			0xFF,
			0x00
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"amp decay",
			"Amp env decay time",
			0x00,
			0xFF,
			0x00
		},
	
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"amp sustain",
			"Amp env sustain level",
			0x00,
			0xFF,
			0xFF
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"amp release",
			"Amp env release time",
			0x00,
			0xFF,	
			0x00
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_OPTION,
			"filter type",
			"Filter type",
			0,
			3,
			0
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"cutoff",
			"Filter cutoff frequency",
			0x00,
			0xFF,
			0xFF
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"resonance",
			"Filter resonance amount",
			0x00,
			0xFF,
			0x00
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"filter env amt",
			"Filter envelope amount",
			0x00,
			0xFF,
			0x00
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"filt env attack",
			"Filter env attack time",
			0x00,
			0xFF,
			0x00
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"filt env decay",
			"Filter env decay time",
			0x00,
			0xFF,
			0x00
		},
	
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"filt env sustain",
			"Filter env sustain level",
			0x00,
			0xFF,
			0x00
		},
		
		{
			OX_FORMAT_SMALL,
			OX_WIDGET_SLIDER,
			"filt env release",
			"Filter env release time",
			0x00,
			0xFF,	
			0x00
		},
		
		{
			OX_FORMAT_TRIGGER,
			OX_WIDGET_BUTTON,
			"gate",
			"Toggle gate",
			0,
			1,
			0
		}
	};
	
	t->long_name		= "Serf SimpleSynth";
	t->short_name		= "simple";
	t->max_tracks		= 1;
	t->input_channels	= 0;
	t->output_channels	= 1;
	t->num_params		= my_num_params;
	t->param_specs		= my_params;
	
	return(1);
}

void SimpleSynth::describe(char* dest, int which, param value) {
	float x;
	int percent;
	
	switch(which) {
	case ix_note:
		sprintf(dest, "$");
		break;
		
	case ix_vol:
		x = ((float)value) / 255.0;
		percent = (int)(x*100);
		sprintf(dest, "%d%%", percent);
		break;
	
	case ix_osc1_wave:
		sprintf(dest, Oscillator::get_wave_name((int)value));
		break;
		
	case ix_osc2_wave:
		sprintf(dest, Oscillator::get_wave_name((int)value));
		break;
		
	case ix_osc1_detune:
		sprintf(dest, "%d cents", ((int)value) - 100);
		break;
		
	case ix_osc2_detune:
		sprintf(dest, "%d cents", ((int)value) - 100);
		break;
		
	case ix_osc_mix:
		x = ((float)value) / 255.0;
		percent = (int)(x * 100);
		sprintf(dest, "%d%% Osc 1 / %d%% Osc 2", 100-percent, percent);
		break;
	
	case ix_amp_attack:
		x = ((float)value) / 255.0;
		sprintf(dest, "%3.3g s", (float)Envelope::value_to_time((samp)x) );
		break;
		
	case ix_amp_decay:
		x = ((float)value) / 255.0;
		sprintf(dest, "%3.3g s", (float)Envelope::value_to_time((samp)x) );
		break;
		
	case ix_amp_sustain:
		x = ((float)value) / 255.0;
		percent = (int)(x*100);
		sprintf(dest, "%d%%", percent);
		break;
		
	case ix_amp_release:
		x = ((float)value) / 255.0;
		sprintf(dest, "%3.3g s", (float)Envelope::value_to_time((samp)x) );
		break;

	case ix_filter_type:
		sprintf(dest, Filter::get_type_name((int)value));
		break;
		
	case ix_filter_cutoff:
		x = ((float)value) / 255.0;
		sprintf(dest, "%i Hz", (int)value_to_hz((freq)x) );
		break;
		
	case ix_filter_resonance:
		x = ((float)value) / 255.0;
		percent = (int)(x*100);
		sprintf(dest, "%d%%", percent);
		break;

	case ix_filter_env_amount:
		x = ((float)value) / 255.0;
		percent = (int)(x*100);
		sprintf(dest, "%d%%", percent);
		break;
		
	case ix_filter_env_attack:
		x = ((float)value) / 255.0;
		sprintf(dest, "%3.3g s", (float)Envelope::value_to_time((samp)x) );
		break;
		
	case ix_filter_env_decay:
		x = ((float)value) / 255.0;
		sprintf(dest, "%3.3g s", (float)Envelope::value_to_time((samp)x) );
		break;
		
	case ix_filter_env_sustain:
		x = ((float)value) / 255.0;
		percent = (int)(x*100);
		sprintf(dest, "%d%%", percent);
		break;
		
	case ix_filter_env_release:
		x = ((float)value) / 255.0;
		sprintf(dest, "%3.3g s", (float)Envelope::value_to_time((samp)x) );
		break;

	case ix_gate:
		sprintf(dest, "Gate toggled.");
		break;
	
	default:
		sprintf(dest, "ERROR");
	}
}

void SimpleSynth::update(int track, int which, param value) {
	freq f;

	switch(which) {
	case ix_note:
		f = m->pkg->note2freq(value);
		osc1.set_pitch(f);
		osc2.set_pitch(f);
		break;
		
	case ix_vol:
		vol = ((samp)value) / 255.0;
		break;
	
	case ix_osc1_wave:
		osc1.set_wave((int)value);
		break;
		
	case ix_osc2_wave:
		osc2.set_wave((int)value);
		break;
		
	case ix_osc1_detune:
		osc1.set_detune_ratio(cents_to_ratio((int)value));
		break;
		
	case ix_osc2_detune:
		osc2.set_detune_ratio(cents_to_ratio((int)value));
		break;
		
	case ix_osc_mix:
		osc_mix = ((samp)value) / 255.0;
		break;
	
	case ix_amp_attack:
		amp_env.set_attack(((samp)value) / 255.0);
		break;
		
	case ix_amp_decay:
		amp_env.set_decay(((samp)value) / 255.0);
		break;
		
	case ix_amp_sustain:
		amp_env.set_sustain(((samp)value) / 255.0);
		break;
		
	case ix_amp_release:
		amp_env.set_release(((samp)value) / 255.0);
		break;
	
	case ix_filter_type:
		filter.set_type((int)value);
		break;
		
	case ix_filter_cutoff:
		filter.set_cutoff(value_to_hz(((freq)value) / 255.0));
		break;
		
	case ix_filter_resonance:
		filter.set_resonance(((freq)value) / 255.0);
		break;
	
	case ix_filter_env_amount:
		filter_env_amount = ((samp)value) / 255.0;
		break;
		
	case ix_filter_env_attack:
		filter_env.set_attack(((samp)value) / 255.0);
		break;
		
	case ix_filter_env_decay:
		filter_env.set_decay(((samp)value) / 255.0);
		break;
		
	case ix_filter_env_sustain:
		filter_env.set_sustain(((samp)value) / 255.0);
		break;
		
	case ix_filter_env_release:
		filter_env.set_release(((samp)value) / 255.0);
		break;
	
	case ix_gate:
		if(gate){
			amp_env.gate_on();
			filter_env.gate_on();
		} else {
			amp_env.gate_off();
			filter_env.gate_off();
		}
		gate = gate ^ 1; // flip it
		break;
	}
}

int SimpleSynth::work(int block_size) {
	samp out;
	
	if(!amp_env.get_stage()) return 0;
	
	for (int i=0; i < block_size; i++) {
		// init
		out = 0.0;
		
		// osc
		out += (1.0 - osc_mix) * osc1.next_samp();
		out += osc_mix * osc2.next_samp();
		
		// filter
		if((i & 0x1f) == 0) {	// only recompute coefs every 32 samples
			filter.set_cutoff_offset(value_to_hz(filter_env.next_samp() * filter_env_amount));
			filter.compute_coefs();
		} else filter_env.next_samp();
		out = filter.next_samp(out);
		
		// amp
		out *= amp_env.next_samp();
		
		// out
		lout[i] = out * vol;
	}

	return(1);
}

int SimpleSynth::track(int change) {
	return(0);  // Not polyphonic yet.
}

inline freq SimpleSynth::cents_to_ratio(int cents) {
/*
	returns the value a pitch should be multiplied by to detune it
	by the given number of cents.
*/
	return (freq)pow( 2.0, ( ((double)(cents - 100)) / 1200.0 ) );
}

inline freq SimpleSynth::value_to_hz(freq c) {
	if (c < 0) c = 0;
	if (c > 1) c = 1;
	return (freq)(20.0 * pow(1000.0, c));	// scales from 20-20000 Hz
}
