/* simplesynth.h
 * ------------------------------------------------------------------------
 *
 * Copyright (C) 2000 Matthew Stanton <matthews@softhome.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 * ------------------------------------------------------------------------
 *
 */

#include <math.h>
#include <stdio.h>

#define SAMPLE_RATE 44100.0
#ifndef PI
#define PI M_PI
#endif

class Oscillator {
private:
	// if the line below changes, also change Oscillator::get_wave_name
	// and Oscillator::set_wave
	enum {saw = 0, pulse = 1, triangle = 2, sine = 3} waveform;
	freq phase;
	freq phaseinc;
	freq pitch;			// Hz
	freq detune_ratio;
	
	inline void update_pitch();
public:
	Oscillator();
	~Oscillator();
	inline samp fast_sin(samp x);
	void set_wave(int w);
	static char* get_wave_name(int w);
	void set_pitch(freq p);
	inline void set_detune_ratio(freq d);
	inline samp next_samp();
};

class Envelope {
private:
	enum {off = 0, attack = 1, decay = 2, sustain = 3, release = 4} stage;
	samp a_step; // time
	samp d_step; // time
	samp s_level; // level
	samp r_step; // time
	samp level;
	
	inline samp value_to_step(samp x);
public:
	Envelope();
	~Envelope();
	void set_attack(samp a);
	void set_decay(samp d);
	void set_sustain(samp s);
	void set_release(samp r);
	void gate_on();
	void gate_off();
	int get_stage();
	inline samp next_samp();
	
	inline static samp value_to_time(samp x);
};

class Filter {
private:
	freq cutoff;		// in Hz
	freq cutoff_offset;	// also in Hz
	freq resonance;		// 0.0 - 1.0, scaling from Makk M4
	enum {lowpass = 0, highpass = 1, bandpass = 2, bandreject = 3} type;
	samp *coefs;
	samp *old_in;
	samp *old_out;
	
	inline freq res_to_q(freq r);
	inline freq res_to_bw(freq r);
	void clear();
public:
	Filter();
	~Filter();
	void set_cutoff(freq c);
	void set_cutoff_offset(freq c);
	void set_resonance(freq r);
	void set_type(int t);
	static char* get_type_name(int t);
	void compute_coefs();
	inline samp next_samp(samp s);
};

class SimpleSynth : public OX_Machine {
private:
	// my_params has moved to SimpleSynth::initialize
	enum {
		ix_note					= 0,
		ix_vol					= 1,
		ix_osc1_wave			= 2,
		ix_osc2_wave			= 3,
		ix_osc1_detune			= 4,
		ix_osc2_detune			= 5,
		ix_osc_mix				= 6,
		ix_amp_attack			= 7,
		ix_amp_decay			= 8,
		ix_amp_sustain			= 9,
		ix_amp_release			= 10,
		ix_filter_type			= 11,
		ix_filter_cutoff		= 12,
		ix_filter_resonance		= 13,
		ix_filter_env_amount	= 14,
		ix_filter_env_attack	= 15,
		ix_filter_env_decay		= 16,
		ix_filter_env_sustain	= 17,
		ix_filter_env_release	= 18,
		ix_gate					= 19
	};
	enum {my_num_params = 20};
	enum {my_track = 0};
	samp vol;
	samp osc_mix;
	Oscillator osc1;
	Oscillator osc2;
	Filter filter;
	Envelope filter_env;
	samp filter_env_amount;
	Envelope amp_env;
	int gate;
	
	inline freq cents_to_ratio(int cents);
	inline static freq value_to_hz(freq c);
public:
	SimpleSynth(machine *m);
	~SimpleSynth();
	static int SimpleSynth::initialize(machine_type* t);
	static void SimpleSynth::describe(char* dest, int which, param value);
	void SimpleSynth::update(int track, int which, param value);
	int SimpleSynth::work(int block_size);
	int SimpleSynth::track(int change);
};
